import amqpstorm

from functools import partial
from injector import Injector, inject
from limecore.messaging.api import (
    Handler,
    Message,
    RollbackException,
    RollbackStrategy,
    Subscriber as _Subscriber,
    Subscription,
)
from logging import Logger
from typing import Callable, List

from .configuration import Configuration


class Subscriber(_Subscriber):
    @inject
    def __init__(
        self,
        configuration: Configuration,
        get_logger: Callable[[str], Logger],
        injector: Injector,
    ):
        self._configuration = configuration
        self._injector = injector
        self._logger = get_logger(self.__class__.__name__)
        self._subscriptions: List[Subscription] = []

    def add_subscription(self, subscription: Subscription):
        self._subscriptions.append(subscription)

        return self

    def run(self):
        self._logger.info("Starting RabbitMQ Subscriber.")

        with amqpstorm.Connection(
            self._configuration.host,
            self._configuration.user,
            self._configuration.password,
        ) as connection:
            with connection.channel() as channel:
                self._logger.info(
                    "Declaring %s Exchange." % self._configuration.exchange
                )
                # Ensure the Exchange exists.
                channel.exchange.declare(
                    self._configuration.exchange, exchange_type="direct"
                )

                # Set QoS to limit number of messages that the consumer will prefetch
                # from the remote.
                channel.basic.qos(self._configuration.prefetch_count)

                # Set up the Subscriptions.
                for subscription in self._subscriptions:
                    self._logger.info("Declaring %s Queue." % subscription.queue)
                    # Ensure the Queue exists.
                    channel.queue.declare(subscription.queue, durable=True)

                    self._logger.info(
                        "Binding queue %s to %s, on %s."
                        % (
                            subscription.queue,
                            subscription.routing_key,
                            self._configuration.exchange,
                        )
                    )
                    # Bind the Queue to the Exchange for the Routing Key.
                    channel.queue.bind(
                        subscription.queue,
                        exchange=self._configuration.exchange,
                        routing_key=subscription.routing_key,
                    )

                    self._logger.info(
                        "Adding a Subscription for %s." % (subscription.queue)
                    )
                    # Configure our subscription to consume from the Queue
                    channel.basic.consume(
                        partial(
                            self._on_message, subscription.handler, subscription.queue
                        ),
                        subscription.queue,
                        no_ack=False,
                    )

                self._logger.info("Starting Consumer.")
                # Start consuming the queue, using the on_message callback.
                channel.start_consuming()

    def _on_message(self, handler: type, queue: str, msg: amqpstorm.Message):
        self._logger.info("Received Message from %s." % queue)

        payload = msg.body
        # If the message is of a known Content-Type, decode it automatically.
        if msg.content_type == "application/json":
            payload = Message.from_json(msg.body)

        self._logger.debug("Message Content: %s" % str(payload))

        try:
            self._logger.debug("Dispatching message to %s." % handler.__name__)

            self._logger.info(self._injector)
            impl = self._injector.get(handler)

            assert isinstance(impl, Handler)

            impl.handle_message(payload)

            self._logger.info(
                "%s completed without error, ACKing message." % handler.__name__
            )

            msg.ack()
        except RollbackException as e:
            self._logger.info("%s failed: %s." % (handler.__name__, str(e)))

            self._logger.info("%s failed, Rolling back..." % handler.__name__)

            if e.rollback_strategy == RollbackStrategy.RETRY_NOW:
                msg.nack(requeue=True)
            elif e.rollback_strategy == RollbackStrategy.RETRY_LATER:
                msg.reject(requeue=True)
            elif e.rollback_strategy == RollbackStrategy.QUARANTINE:
                raise NotImplementedError("RollbackStrategy.QUARANTINE")
            else:
                raise Exception("no rollback stragy")
        except Exception as e:
            self._logger.info("%s failed: %s." % (handler.__name__, str(e)))

            msg.reject(requeue=True)



from injector import Module as _Module, provider, singleton
from limecore.core.configuration import Configuration
from limecore.messaging.api import Publisher, Subscriber

from .configuration import Configuration as RabbitMQConfiguration
from .publisher import Publisher as RabbitMQPublisher
from .subscriber import Subscriber as RabbitMQSubscriber


class Module(_Module):
    @singleton
    @provider
    def provide_configuration(
        self, configuration: Configuration
    ) -> RabbitMQConfiguration:
        rabbitmq_configuration = configuration.section(
            "limecore", "messaging", "rabbitmq"
        )

        return RabbitMQConfiguration(
            exchange=rabbitmq_configuration.get_string("exchange") or "amqp",
            host=rabbitmq_configuration.get_string("host") or "localhost",
            password=rabbitmq_configuration.get_string("password") or "guest",
            prefetch_count=rabbitmq_configuration.get_string("prefetch_count") or 2,
            user=rabbitmq_configuration.get_string("password") or "guest",
        )

    @provider
    def provide_publisher(self) -> Publisher:
        return self.__injector__.create_object(RabbitMQPublisher)

    @provider
    def provide_subscriber(self) -> Subscriber:
        return self.__injector__.create_object(RabbitMQSubscriber)



from .configuration import Configuration


class ConfigurationBuilder:
    def _init_(self):
        self._exchange = "amqp"
        self._host = "localhost"
        self._password = "guest"
        self._prefetch_count = 2
        self._user = "guest"

    @staticmethod
    def create():
        return ConfigurationBuilder()

    def build(self) -> dict:
        return Configuration(
            self._exchange, self._host, self._password, self._user, self._prefetch_count
        )

    def with_exchange(self, exchange: str):
        self._exchange = exchange

        return self

    def with_host(self, host: str):
        self._host = host

        return self

    def with_password(self, password: str):
        self._password = password

        return self

    def with_prefetch_count(self, prefetch_count: int):
        self._prefetch_count = prefetch_count

        return self

    def with_user(self, user: str):
        self._user = user

        return self

    def _str_(self):
        return "%s@%s/%s" % (self._user, self._host, self._exchange)



class Configuration:
    def __init__(
        self,
        exchange: str = "amqp",
        host: str = "localhost",
        password: str = "guest",
        user: str = "guest",
        prefetch_count: int = 2,
    ):
        self.exchange = exchange
        self.host = host
        self.password = password
        self.prefetch_count = prefetch_count
        self.user = user



import amqpstorm

from injector import inject
from limecore.messaging.api import Message, Publisher as _Publisher

from .configuration import Configuration


class Publisher(_Publisher):
    @inject
    def __init__(self, configuration: Configuration):
        self._configuration = configuration

        self._channel = None
        self._connection = None

    def __enter__(self):
        self._connection = amqpstorm.Connection(
            self._configuration.host,
            self._configuration.user,
            self._configuration.password,
        ).__enter__()
        self._channel = self._connection.channel().__enter__()

        return self

    def __exit__(self, *args, **kwargs):
        self._channel.__exit__(*args, **kwargs)
        self._connection.__exit__(*args, **kwargs)

    def publish(self, msg: Message, mandatory: bool = True, queue: str = None):
        assert self._channel is not None

        if queue is not None:
            # If a named Queue is specified, declare it to ensure it exist.
            self._channel.queue.declare(queue, durable=True)
        else:
            # If no named Quue is specified, declare the default exchange which
            # will route the message.
            self._channel.exchange.declare(
                self._configuration.exchange, exchange_type="direct"
            )

        # Create the Message Properties
        properties = {
            "content_type": "application/json",
        }

        # Create the actual Message
        message = amqpstorm.Message.create(self._channel, msg.to_json(), properties)

        # Publish the Message to the Queue
        message.publish(
            exchange=queue is None and self._configuration.exchange or "",
            mandatory=mandatory,
            routing_key=queue is None and msg.routing_key or queue,
        )
